"""Compute minimal trap-spaces of a Petri-net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr and giang.trinh91@gmail.com

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import json
import os
import subprocess
import sys
import tempfile
import xml.etree.ElementTree as etree
from typing import Generator, IO, List, Set

import networkx as nx  # TODO maybe replace with lists/dicts

from . import pnml_to_asp, version
from .bnet import read_bnet
from .cp import get_cp_solutions
from .max_sat import get_sat_solutions
from .naive import write_naive_asp


def read_pnml(fileobj: IO) -> nx.DiGraph:
    """Parse the given file."""
    root = etree.parse(fileobj).getroot()
    if root.tag != "pnml":
        raise ValueError("Currently limited to parsing PNML files")
    net = nx.DiGraph()

    for place in root.findall("./net/place"):
        net.add_node(
            place.get("id"), kind="place"  # , name=place.find("./name/value").text
        )

    for transition in root.findall("./net/transition"):
        net.add_node(transition.get("id"), kind="transition")

    for arc in root.findall("./net/arc"):
        net.add_edge(arc.get("source"), arc.get("target"))

    return net


def write_asp(petri_net: nx.DiGraph, asp_file: IO):
    """Write the ASP program for the maximal conflict-free siphons of petri_net."""
    for node, kind in petri_net.nodes(data="kind"):
        if kind == "place":
            print("{", pnml_to_asp(node), "}.", file=asp_file, sep="")
            if not node.startswith("-"):
                print(
                    f":- {pnml_to_asp(node)}, {pnml_to_asp('-' + node)}.", file=asp_file
                )  # conflict-freeness
        # it's a transition, apply siphon
        # (if one succ is true, one pred must be true)
        else:
            preds = list(petri_net.predecessors(node))
            or_preds = "; ".join(map(pnml_to_asp, preds))
            for succ in petri_net.successors(node):
                if succ not in preds:  # optimize obvious tautologies
                    print(f"{or_preds} :- {pnml_to_asp(succ)}.", file=asp_file)


def solve_asp(asp_filename: str, max_output: int, time_limit: int) -> str:
    """Run an ASP solver on program asp_file and get the solutions."""
    result = subprocess.run(
        [
            "clingo",
            str(max_output),
            # TODO try clasp parrallel-mode f"--parallel-mode={os.cpu_count()}",
            "--heuristic=Domain",  # maximal w.r.t. inclusion
            "--enum-mod=domRec",
            "--dom-mod=3,16",
            "--outf=2",  # json output
            f"--time-limit={time_limit}",
            asp_filename,
        ],
        capture_output=True,
        text=True,
    )

    # https://www.mat.unical.it/aspcomp2013/files/aspoutput.txt
    # 30: SAT, all enumerated, optima found, 10 stopped by max
    if result.returncode != 30 and result.returncode != 10:
        print(f"Return code from clingo: {result.returncode}")
        result.check_returncode()  # will raise CalledProcessError

    return result.stdout


def solution_to_bool(places: List[str], sol: Set[str]) -> List[str]:
    """Convert a list of present places in sol, to a tri-valued vector."""
    return [place_in_sol(sol, p) for p in places]


def place_in_sol(sol: Set[str], place: str) -> str:
    """Return 0/1/- if place is absent, present or does not appear in sol.

    Remember that being in the siphon means staying empty,
    so the opposite value is the one fixed.
    """
    if "p" + place in sol:
        return "0"
    if "n" + place in sol:
        return "1"
    return "-"


def get_solutions(
    asp_output: str, places: List[str]
) -> Generator[List[str], None, None]:
    """Display the ASP output back as trap-spaces."""
    solutions = json.loads(asp_output)
    yield from (
        solution_to_bool(places, set(sol["Value"]))
        for sol in solutions["Call"][0]["Witnesses"]
    )


def get_asp_output(
    petri_net: nx.DiGraph,
    max_output: int,
    time_limit: int,
    method: str,
    debug: bool,
    nprocs: int,
) -> str:
    """Generate and solve ASP file."""
    (_, tmpname) = tempfile.mkstemp(suffix=".lp", text=True)
    with open(tmpname, "wt") as asp_file:
        if method == "asp":
            write_asp(petri_net, asp_file)
        elif method == "naive":
            write_naive_asp(petri_net, asp_file, nprocs)
    if debug:
        print(f"ASP file {tmpname} written.")
    solutions = solve_asp(tmpname, max_output, time_limit)
    if not debug:
        os.unlink(tmpname)
    return solutions


def compute_trap_spaces(
    infile: IO,
    display: bool = False,
    max_output: int = 0,
    time_limit: int = 0,
    method: str = "asp",
    debug: bool = False,
    nprocs: int = 1,
) -> Generator[List[str], None, None]:
    """Do the minimal trap-space computation on input file infile."""
    toclose = False
    if isinstance(infile, str):
        infile = open(infile, "r", encoding="utf-8")
        toclose = True

    if infile.name.endswith(".pnml") and method in ("asp", "sat", "cp"):
        petri_net = read_pnml(infile)
    elif infile.name.endswith(".bnet"):
        petri_net = read_bnet(infile, method)
    else:
        infile.close()
        raise ValueError("Failed parsing input")

    if toclose:
        infile.close()

    if debug:
        print("Input file parsed.")

    places = []
    for node, kind in petri_net.nodes(data="kind"):
        if kind == "place" and not node.startswith("-"):
            places.append(node)
    if display:
        print(" ".join(places))

    if method == "sat":
        solutions = get_sat_solutions(petri_net, max_output, time_limit, places)
    elif method == "cp":
        solutions = get_cp_solutions(petri_net, max_output, time_limit, places, nprocs)
    else:
        solutions_output = get_asp_output(
            petri_net, max_output, time_limit, method, debug, nprocs
        )
        if debug:
            print("ASP solutions obtained.")
        solutions = get_solutions(solutions_output, places)

    if display:
        print("\n".join(" ".join(sol) for sol in solutions))
        return
    else:
        yield from solutions


def main():
    """Read the Petri-net send the output to ASP and print solution."""
    parser = argparse.ArgumentParser(
        description=" ".join(__doc__.splitlines()[:3]) + " GPLv3"
    )
    parser.add_argument(
        "-d",
        "--debug",
        action=argparse.store_true,
        help="Print debugging information.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s v{version}".format(version=version),
    )
    parser.add_argument(
        "-m",
        "--max",
        type=int,
        default=0,
        help="Maximum number of solutions (0 for all).",
    )
    parser.add_argument(
        "-p",
        "--parallel",
        type=int,
        default=1,
        help="Maximum number of cores to use [only for naive method] (0 for no-limit).",
    )
    parser.add_argument(
        "-t",
        "--time",
        type=int,
        default=0,
        help="Maximum number of seconds for search (0 for no-limit).",
    )
    parser.add_argument(
        "-s",
        "--solver",
        choices=["asp", "cp", "sat", "naive"],
        default="asp",
        type=str,
        help="Solver to compute the Maximal conflict-free siphons.\n"
             "'asp' requires clingo, 'cp' requires minizinc.",
    )
    parser.add_argument(
        "infile",
        type=argparse.FileType("r", encoding="utf-8"),
        nargs="?",
        default=sys.stdin,
        help="Petri-net (PNML) file",
    )
    args = parser.parse_args()

    try:
        next(
            compute_trap_spaces(
                args.infile,
                display=True,
                max_output=args.max,
                time_limit=args.time,
                method=args.solver,
                debug=args.debug,
                nprocs=args.parallel,
            )
        )
    except StopIteration:
        pass


if __name__ == "__main__":
    main()



"""Compute minimal trap-spaces of a Petri-net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from itertools import islice
from typing import Generator, List

import networkx as nx  # TODO maybe replace with lists/dicts

from pysat.examples.rc2 import RC2
from pysat.formula import WCNF


def create_maxsat(petri_net: nx.DiGraph) -> WCNF:
    """Create the MaxSAT program for the maximal conflict-free siphons of petri_net."""
    variables = {}
    wcnf = WCNF()
    for i, node in enumerate(
        (n for n, k in petri_net.nodes(data="kind") if k == "place"), start=1
    ):
        variables[node] = i
        wcnf.append([i], weight=1)
    for node, kind in petri_net.nodes(data="kind"):
        if kind == "place":
            if not node.startswith("-"):
                wcnf.append(
                    [-variables[node], -variables["-" + node]]
                )  # conflict-freeness
        else:  # it's a transition, apply siphon (if one succ is true, one pred must be true)
            or_preds = [variables[p] for p in petri_net.predecessors(node)]
            for succ in petri_net.successors(node):
                vsucc = variables[succ]
                if vsucc not in or_preds:  # optimize obvious tautologies
                    wcnf.append(or_preds + [-vsucc])
    return wcnf


def solve_maxsat(
    cnf: WCNF, max_output: int, time_limit: int
) -> Generator[List[int], None, None]:
    """Run an MaxSAT solver on given WCNF and get the solutions."""
    # print(cnf.hard)
    rc2 = RC2(cnf)
    sol = rc2.enumerate(block=-1)  # minimal correction subsets
    yield from islice(sol, None if max_output == 0 else max_output)


def sat_to_bool(
    places: List[str], all_places: List[str], model: List[int]
) -> List[str]:
    """Transform a model to a list of 0/1/- strings."""
    result = []
    for place in places:
        if all_places.index(place) + 1 in model:
            result.append("0")
        elif all_places.index("-" + place) + 1 in model:
            result.append("1")
        else:
            result.append("-")
    return result


def get_sat_solutions(
    petri_net: nx.DiGraph, max_output: int, time_limit: int, places: List[str]
) -> Generator[List[str], None, None]:
    """Print the solutions."""
    all_places = [
        node for node, kind in petri_net.nodes(data="kind") if kind == "place"
    ]
    for model in solve_maxsat(create_maxsat(petri_net), max_output, time_limit):
        yield sat_to_bool(places, all_places, model)



"""Compute minimal trap-spaces of a Petri-net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from math import ceil
from multiprocessing import Pool, cpu_count, current_process
from os import unlink
from sys import setrecursionlimit
from typing import IO, List, Set, Tuple

import networkx as nx  # TODO maybe replace with lists/dicts

from pyeda.boolalg.bdd import bdd2expr, expr2bdd
from pyeda.boolalg.expr import And, AndOp, Constant, Literal, Or, OrOp, Variable, expr

# from pyeda.boolalg.minimization import espresso_exprs

from . import pnml_to_asp


def write_naive_asp(petri_net: nx.DiGraph, asp_file: IO, nprocs: int):
    """Write the ASP program for naive encoding of trap spaces."""
    nnodes = petri_net.number_of_nodes()
    if nprocs == 0:
        nproc = cpu_count()
    else:
        nproc = min(nprocs, cpu_count())
    if nproc > 1:
        chunksize = ceil(nnodes / (8 * nproc))  # seems decent
        with Pool(nproc, setup_worker, (asp_file.name,)) as p:
            pids = set(
                p.imap_unordered(
                    add_variable,
                    petri_net.nodes(data=True),
                    chunksize,
                )
            )
        for p in pids:
            with open(f"{asp_file.name}_{p}", "r") as f:
                for line in f:
                    asp_file.write(line)
            unlink(f"{asp_file.name}_{p}")
    else:
        setrecursionlimit(2048)
        globals()["counter"] = 0
        globals()["pid"] = 0
        globals()["asp_file"] = asp_file
        for node_and_data in petri_net.nodes(data=True):
            add_variable(node_and_data)


def setup_worker(filename):
    """Set global variables for subprocess."""
    # big bnets…
    setrecursionlimit(2048)
    global counter
    counter = 0
    global pid
    pid = current_process().pid
    while pid is None:
        pid = current_process().pid
    global asp_file
    asp_file = open(f"{filename}_{pid}", "wt")


def add_variable(node_and_data):
    """Add all the rules for one variable."""
    node, data = node_and_data
    name = pnml_to_asp(node)
    print("{", name, "}.", sep="", file=asp_file)
    print(f"#show {name}/0.", file=asp_file)
    if not node.startswith("-"):
        print(
            f":- {name}, {pnml_to_asp('-' + node)}.", file=asp_file
        )  # conflict-freeness
    add_tree(expr(data["function"]).to_nnf(), expr(data["var"]), asp_file)
    asp_file.flush()
    return pid


def add_tree(source: expr, target: expr, asp_file):
    """Add the AST of things->target to the ASP program."""
    global counter
    if isinstance(target, Variable) and target.name.startswith("aux"):
        starget = target.name
    else:
        starget = pnml_to_asp(str(~target))

    if isinstance(source, Literal):
        print(f"{pnml_to_asp(str(~source))} :- {starget}.", file=asp_file)
    elif isinstance(source, Constant):
        if source.is_zero():
            print(f":- {starget}.", file=asp_file)
        elif source.is_one():
            # nothing to do
            pass
        else:
            raise ValueError(f"Houston we have a problem with {source}…")
    elif isinstance(source, OrOp):
        safe, unsafe = split_safe_unsafe(source)
        if unsafe:
            source = Or(cnf_from_bdd(Or(*unsafe)), *safe)
            if not isinstance(source, OrOp):
                return add_tree(source, target, asp_file)
        source_str = ""
        for s in source.xs:
            if isinstance(s, Literal):
                svs = pnml_to_asp(str(~s))
            else:
                vs = expr(f"aux_{pid}_{counter}")
                counter += 1
                add_tree(s, vs, asp_file)
                svs = str(vs)
            if source_str:
                source_str += "; " + svs
            else:
                source_str = svs
        print(f"{source_str} :- {starget}.", file=asp_file)
    elif isinstance(source, AndOp):
        for s in source.xs:
            if isinstance(s, Literal):
                print(f"{pnml_to_asp(str(~s))} :- {starget}.", file=asp_file)
            else:
                add_tree(s, target, asp_file)
    else:
        print(f"Houston we have a problem with {source}…")


def leaves(expression: expr) -> Set[Literal]:
    """Return all the Litterals in expression."""
    # TODO cache?
    s = set()
    for ex in expression.iter_dfs():
        if isinstance(ex, Literal):
            s.add(ex)
    return s


def split_safe_unsafe(expression: expr) -> Tuple[List[expr], List[expr]]:
    """Split an Or into safe and unsafe (contains a literal and its negation) parts."""
    assert isinstance(expression, OrOp)
    leaves_list = [leaves(child) for child in expression.xs]
    conflicts = set()
    for child1, child2 in [
        (a, b) for index, a in enumerate(leaves_list) for b in leaves_list[index + 1 :]
    ]:
        for v in child1:
            if ~v in child2:
                conflicts.add(v)
                conflicts.add(~v)
    safe = []
    unsafe = []
    for child, child_leaves in zip(expression.xs, leaves_list):
        if child_leaves & conflicts:
            unsafe.append(child)
        else:
            safe.append(child)
    return (safe, unsafe)


def cnf_from_bdd(source):
    """Get a CNF via a BDD."""
    conjunct = []
    for success in expr2bdd(~source).satisfy_all():
        disjunct = []
        for p, v in success.items():
            if v == 0:
                disjunct.append(bdd2expr(p))
            else:
                disjunct.append(~bdd2expr(p))
        conjunct.append(Or(*disjunct))
    result = And(*conjunct)
    return result



"""Compute minimal trap-spaces of a Petri-net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from dataclasses import asdict
from datetime import timedelta
from time import perf_counter
from typing import Generator, List

from minizinc import Instance, Model, Result, Solver, Status

import networkx as nx  # TODO maybe replace with lists/dicts


def create_cp(petri_net: nx.DiGraph) -> Model:
    """Create the CP program for the max conflict-free siphons of petri_net."""
    model = Model()
    model.add_string('include "globals.mzn";\n')
    variables = []
    for node in (n for n, k in petri_net.nodes(data="kind") if k == "place"):
        v = zincify(node)
        variables.append(v)
        model.add_string(f"var bool: {v};\n")
    for node, kind in petri_net.nodes(data="kind"):
        if kind == "place":
            if not node.startswith("-"):
                # conflict-freeness
                model.add_string(f"constraint not {node} \\/ not not_{node};\n")
        else:
            # it's a transition, apply siphon
            # (if one succ is true, one pred must be true)
            or_preds = [zincify(p) for p in petri_net.predecessors(node)]
            for succ in petri_net.successors(node):
                zsucc = zincify(succ)
                if zsucc not in or_preds:  # optimize obvious tautologies
                    or_string = " \\/ ".join(or_preds)
                    model.add_string(f"constraint {zsucc} -> {or_string};\n")
    model.add_string(
        f"solve :: bool_search([{', '.join(variables)}], input_order, indomain_max) satisfy;\n"
    )
    return model


def zincify(variable: str) -> str:
    """Return the name of the corresponding CP variable."""
    if variable.startswith("-"):
        return f"not_{variable[1:]}"
    return variable


def solve_cp(
    model: Model,
    max_output: int,
    time_limit: int,
    nprocs: int,
) -> Generator[List[int], None, None]:
    """Run an CP solver on given WCNF and get the solutions."""
    solver = Solver.lookup("gecode")
    inst = Instance(solver, model)
    nsol = 0
    if time_limit > 0:
        remaining = timedelta(seconds=time_limit)
    else:
        remaining = None
    # TODO handle time limit?
    while (remaining is None or remaining > timedelta()) and \
          (max_output == 0 or nsol < max_output):
        start = perf_counter()
        result = inst.solve(timeout=remaining, processes=nprocs)
        end = perf_counter()
        if remaining is not None:
            remaining -= timedelta(seconds=(end - start))
            # print(f"{remaining} remaining")
        if result.status == Status.SATISFIED:
            nsol += 1
            # print(result.solution)
            yield result
            d = asdict(result.solution)
            del d["_checker"]
            # lexicographic constraint to restart from last solution
            # first remove old one if there was already a solution found
            if nsol > 1:
                with model._lock:
                    del model._code_fragments[-2]
            model.add_string(
                f"constraint lex_less([{', '.join(d.keys())}], "
                f" [{', '.join(map(str, map(int, d.values())))}]);"
            )
            # non subset constrait for maximality
            non_sub = []
            for k, v in d.items():
                if not v:
                    non_sub.append(k)
            or_string = " \\/ ".join(non_sub)
            model.add_string(f"constraint {or_string};")
            inst = Instance(solver, model)
        else:
            print(result.status)
            break


def cp_to_bool(places: List[str], res: Result) -> List[str]:
    """Transform a Result to a list of 0/1/- strings."""
    result = []
    for place in places:
        if res[place]:
            result.append("0")
        elif res["not_" + place]:
            result.append("1")
        else:
            result.append("-")
    return result


def get_cp_solutions(
    petri_net: nx.DiGraph,
    max_output: int,
    time_limit: int,
    places: List[str],
    nprocs: int,
) -> Generator[List[str], None, None]:
    """Print the solutions."""
    for result in solve_cp(create_cp(petri_net), max_output, time_limit, nprocs):
        yield cp_to_bool(places, result)



"""Compute minimal trap-spaces of a Petri-net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from sys import setrecursionlimit
from typing import IO

import networkx as nx  # TODO maybe replace with lists/dicts

from pyeda.boolalg.bdd import bddvar, expr2bdd
from pyeda.boolalg.expr import expr


def add_edges(net: nx.DiGraph, tname: str, things: expr, source: str):
    """Add all edges from things to tname and back in net except for source."""
    if source.startswith("-"):
        nsource = source[1:]
    else:
        nsource = "-" + source

    for i, t in enumerate(things.satisfy_all()):
        name = f"{tname}_{i}"
        net.add_node(name, kind="transition")
        for p, v in t.items():
            if v == 0:
                pname = "-" + str(p)
            else:
                pname = str(p)
            net.add_edge(pname, name)
            if (
                pname == source
            ):  # don't add an edge back to "source" but add to Not(source)
                net.add_edge(name, nsource)
            else:
                net.add_edge(name, pname)


def read_bnet(fileobj: IO, method: str) -> nx.DiGraph:
    """Parse a BoolNet .bnet file and build the corresponding Petri net."""
    net = nx.DiGraph()

    # big bnets…
    setrecursionlimit(2048)
    for line in fileobj.readlines():
        if line.startswith("#") or line.startswith("targets, factors"):
            continue
        line = line.split("#")[0]
        try:
            x, fx = line.replace(" ", "").replace("!", "~").split(",", maxsplit=1)
        except ValueError:  # not enough values to unpack
            continue
        net.add_node(x, kind="place")
        net.add_node(
            "-" + x, kind="place"
        )  # convention in PNML files obtained from bnet
        if method in ("asp", "sat", "cp"):
            vx = bddvar(x)
            fx = expr2bdd(expr(fx))
            # print(x, fx)
            activate = fx & ~vx
            inactivate = ~fx & vx
            # print(activate, inactivate, sep="\n")

            add_edges(net, f"tp_{x}", activate, "-" + x)
            add_edges(net, f"tn_{x}", inactivate, x)
        else:
            # delay expr parsing to parallel jobs (for pickling)
            net.nodes[x]["function"] = fx
            net.nodes[x]["var"] = x
            net.nodes["-" + x]["function"] = "~(" + fx + ")"
            net.nodes["-" + x]["var"] = "~(" + x + ")"

    return net



# noqa: D104
# pylint: disable=missing-docstring
"""Compute minimal trap spaces of a Petri net encoded Boolean model.

Copyright (C) 2022 Sylvain.Soliman@inria.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# Don't forget to git tag v<version> && git push --tags
version = "0.7.0"  # pylint: disable=invalid-name


def pnml_to_asp(name: str) -> str:
    """Convert a PNML id to an ASP variable."""
    # TODO handle non-accetable chars
    name = name.replace("~", "-")
    if name.startswith("-"):
        return "n" + name[1:]
    return "p" + name


from .trappist import compute_trap_spaces  # noqa
